//================================================================================
// SimpleKickVote for Unreal Tournament 4.36 by James P.
// Version 1.0 (20090222)
// Please give credit if you modify this mutator
//================================================================================

class SimpleKickVote extends Mutator
	config( SimpleKickVote );

// Config attributes
var config bool bDebug;
var config float votePercent, voteTimeout;

// Internal attributes
var private bool bInitialized, bVoteActive;
var private int votedYes, votedNo;
var private Pawn invokingPlayer, kickPlayer;
var private Pawn alreadyVoted[ 32 ]; // Maybe increase size
var private string tempBans[ 32 ]; // Maybe increase size

function PreBeginPlay()
{
	super.PreBeginPlay();

	// Necessary to get round a bug that causes PreBeginPlay() to get called twice
	if( bInitialized )
		return;
	else
		bInitialized = true;

	log( "Starting SimpleKickVote" , 'SimpleKickVote' );
	log( "Please contact the author to submit bugs and suggestions" , 'SimpleKickVote' );

	SaveConfig(); // Save ini settings and create ini file in the same go

	log( "Registering as message mutator" , 'SimpleKickVote' );
	Level.Game.RegisterMessageMutator( self );
}

function Mutate( string mutateString, PlayerPawn pp )
{
	local int i, chosenPID;
	local float newPercent, newTimeout;
	local Pawn p;
	local string voteOption;

	super.Mutate( mutateString, pp );

	if( isValidPlayer( pp ) )
	{
		if( getDebug() )
		{
			log( "mutateString is" @ mutateString );
			log( "PlayerPawn is" @ pp @ "," @ pp.PlayerReplicationInfo.PlayerName );
		}

		if ( !getVoteActive() && Left( Caps( MutateString ), 13 ) ~= "SKV STARTVOTE" )
		{
			if( getDebug() ) log( "mutate skv startvote" );

			setInvokingPlayer( pp );
			messageClient( pp, "Choose the player you want to kick using mutate <PID>" );
			showPlayerList( pp );
		}

		if( !getVoteActive() && pp == invokingPlayer && Left( Caps( MutateString ), 8 ) ~= "SKV KICK" )
		{
			if( getDebug() ) log( "mutate skv kick" );

			chosenPID = int( elem( MutateString, 2, " " ) );

			if( getDebug() ) log( "chosenPID =" @ chosenPID );

			if( chosenPID >= 0 )
			{
				setVoteActive( true );

				addVote( "Yes", pp );

				setKickPlayer( getPawnByPID( chosenPID ) );

				if( getDebug() ) log( "kickPlayer =" @ getKickPlayer() @ "," @ getKickPlayer().PlayerReplicationInfo.PlayerName );

				if( getKickPlayer() != none )
				{
					notifyAllPlayers( "Kickvote has been initiated" );

					notifyAllPlayers( "Do you want to kick" @ getKickPlayer().PlayerReplicationInfo.PlayerName @ "?" );
					notifyAllPlayers( "Enter 'mutate skv Vote Yes' in your console to vote in favour" );
					notifyAllPlayers( "Enter 'mutate skv Vote No' in your console to vote against" );

					notifyAllPlayers( "You have" @ voteTimeout @ "seconds to vote" );

					SetTimer( voteTimeout, false );
				}
				else
					messageClient( pp, "Player was not found" );
			}
			else
				messageClient( pp, "Invalid playerID" );
		}

		// Receive votes
		if( getVoteActive() && Left( Caps( MutateString ), 8 ) ~= "SKV VOTE" )
		{
			voteOption = elem( MutateString, 2, " " );

			if( getDebug() )
			{
				log( "mutate skv vote" );
				log( "Player is" @ pp @ "," @ pp.PlayerReplicationInfo.PlayerName );
				log( "voteOption is" @ voteOption );
			}

			if( !hasVoted( pp ) )
			{
				if( addVote( "Yes", pp ) )
					messageClient( pp, "Your vote has been counted" );
				else
					messageClient( pp, "An error occured" );
			}
			else
				messageClient( pp, "You have already voted" );

			checkVotes( false );
		}

		if ( ( pp == getInvokingPlayer() || pp.bAdmin ) && Left( Caps( MutateString ), 14 ) ~= "SKV CANCELVOTE" )
		{
			if( bDebug ) log( "Cancelling vote." );

			setVoteActive( false );

			messageClient( pp, "You cancelled the vote." );

			notifyAllPlayers( "Kickvote has been cancelled." );

			resetVote();
		}

		if ( pp.bAdmin && Left( Caps( MutateString ), 14 ) ~= "SKV SETPERCENT" )
		{
			if( bDebug ) log( "Setting vote percent." );

			newPercent = float( elem( MutateString, 2, " " ) );

			if( bDebug ) log( "newPercent value is" @ newPercent );

			if( newPercent > 0.0 && newPercent <= 1.0 )
			{
				votePercent = newPercent;

				SaveConfig();

				messageClient( pp, "Vote percent set to" @ votePercent );
			}
			else if( newPercent > 1.0 )
				messageClient( pp, "Percent value is above 100%. Please enter a value between 0.0 and 1.0" );
			else if( newPercent <= 0.0 ) // Changed
				messageClient( pp, "Percent value is below 0. Please enter a value between 0.0 and 1.0" );
			else
				messageClient( pp, "Undetermined error" );
		}

		if ( pp.bAdmin && Left( Caps( MutateString ), 14 ) ~= "SKV SETTIMEOUT" )
		{
			if( getDebug() ) log( "Setting vote timeout" );

			newTimeout = float( elem( MutateString, 2, " " ) );

			if( getDebug() ) log( "newTimeout value is" @ newTimeout );

			if( newTimeout > 0 )
			{
				voteTimeout = newTimeout;

				SaveConfig();

				messageClient( pp, "Vote timeout set to" @ voteTimeout );
			}
			else if( newPercent <= 0 )
				messageClient( pp, "Vote timeout below or equal to 0" );
			else
				messageClient( pp, "Undetermined error" );
		}

		if ( pp.bAdmin && Left( Caps( MutateString ), 12 ) ~= "SKV DEBUG ON" )
		{
			if( bDebug ) log( "Setting debug on" );

			setDebug( true );

			messageClient( pp, "bDebug is now" @ getDebug() );
		}

		if ( pp.bAdmin && Left( Caps( MutateString ), 13 ) ~= "SKV DEBUG OFF" )
		{
			if( bDebug ) log( "Setting debug off" );

			setDebug( false );

			messageClient( pp, "bDebug is now" @ getDebug() );
		}
	}
}

function bool addVote( string vote, Pawn p )
{
	local int i;

	if( vote == "" || p == none )
		return false;

	if( vote ~= "yes" )
		votedYes++;
	else if( vote ~= "no" )
		votedNo++;

	for( i = 0; i < 32; i++ )
	{
		if( alreadyVoted[ i ] == none )
			break;
	}

	if( getDebug() ) log( "addVote(): i =" @ i );

	if( i < 32 )
		alreadyVoted[ i ] = p;

	return true;
}

function checkVotes( bool bExpired )
{
	local int i;
	local string playerName, playerIP;
	local float yesPercent, noPercent;

	if( getDebug() )
	{
		log( "Number of yes votes" @ votedYes );
		log( "Number of no votes" @ votedNo );
		log( "Number of players" @ Level.Game.NumPlayers );
		log( "kickPlayer" @ kickPlayer );
	}

	if( bExpired || ( votedYes + votedNo == Level.Game.NumPlayers ) && kickPlayer != none )
	{
		if( getDebug() ) log( "checkVotes(): ban and kick part" );

		yesPercent = votedYes / Level.Game.NumPlayers;
		noPercent = votedNo / Level.Game.NumPlayers;

		notifyAllPlayers( "Vote results:" );
		notifyAllPlayers( votedYes @ "players voted yes (" $ votedYes*100 $ "%)" );
		notifyAllPlayers( votedNo @ "players voted no (" $ votedNo*100 $ "%)" );

		if( yesPercent >= votePercent )
		{
			playerName = kickPlayer.PlayerReplicationInfo.PlayerName;
			playerIP = getPlayerIP( kickPlayer );

			if( playerIP != "" )
			{
				if( addBan( playerIP ) )
				{
					if( bDebug ) log( "Debug: Temporary ban added. PlayerIP is" @ playerIP );

					for( i = 0; i < 32; i++ )
					{
						if( tempBans[ i ] == "" )
						break;
					}

					if( i < 32 )
						tempBans[ i ] = playerIP;
				}
				else
				{
					log( "An error occured while adding a temporary ban. Please check IP policies" , 'SimpleKickVote' );
					log( "Player IP is" @ playerIP, 'SimpleKickVote' );
				}
			}
			else if( !kickPlayer.IsA( 'Bot' ) )
			{
				log( "Empty player IP ! Playername is" @ playerName, 'SimpleKickVote' );

				return; // Exit
			}
			else if( kickPlayer.IsA( 'Bot' ) && getDebug() )
				log( "Player is a bot. Ignoring" );

			kickPlayer.Destroy();
			notifyAllPlayers( playerName @ "was kicked off the server." );
		}
		else
			notifyAllPlayers( "Insufficient votes in favour. Kickvote cancelled." );

		resetVote(); // Reset voting
	}
	else if( ( votedYes + votedNo ) > Level.Game.NumPlayers )
	{
		log( "[Error] Total votes exceeds number of players", 'SimpleKickVote' );

		resetVote();
	}
	else if( kickPlayer == none )
	{
		log( "[Error] kickPlayer is undefined", 'SimpleKickVote' );

		resetVote();
	}
}

function bool getDebug()
{
	return bDebug;
}

function Pawn getInvokingPlayer()
{
	return invokingPlayer;
}

function Pawn getKickPlayer()
{
	return kickPlayer;
}

function bool getVoteActive()
{
	return bVoteActive;
}

function bool hasVoted( Pawn p )
{
	local int i;

	if( getDebug() )
	{
		log( "Call to hasVoted()" );
		log( "Pawn is" @ p @ "," @ p.PlayerReplicationInfo.PlayerName );
	}

	if( p != none )
	{
		for( i = 0; i < ArrayCount( alreadyVoted ) && alreadyVoted[ i ] != none; i++ )
		{
			if( getDebug() )
			{
				log( "alreadyVoted[" $ i $ "]=" @ alreadyVoted[ i ] );
				log( "p == alreadyVoted[ i ]" @  p == alreadyVoted[ i ] );
			}

			if( p == alreadyVoted[ i ])
				return true;
		}
	}

	return false;
}

function notifyAllPlayers( string message )
{
	if( message != "" )
		BroadcastMessage( "SimpleKickvote:" @ message );
}

function setDebug( bool debugValue )
{
	bDebug = debugValue;
}

function setKickPlayer( Pawn p )
{
	kickPlayer = p;
}

function setInvokingPlayer( Pawn p )
{
	invokingPlayer = p;
}

function setVoteActive( bool b )
{
	bVoteActive = b;
}

function showPlayerList( Pawn aPawn )
{
	local Pawn p;

	if( getDebug() )
	{
		log( "showPlayerList() called" );
		log( "aPawn is" @ aPawn @ "," @ aPawn.PlayerReplicationInfo.PlayerName, 'SimpleKickVote' );
	}

	if( aPawn != none )
	{
		p = Level.PawnList; // Get player list

		while( p != none )
		{
			messageClient( aPawn, "PID:" @ p.PlayerReplicationInfo.PlayerID @ "-" @ p.PlayerReplicationInfo.PlayerName );

			p = p.nextPawn;
		}
	}
}

function messageClient( Pawn p, string message )
{
	if( p != none )
		p.ClientMessage( "SimpleKickvote:" @ message );
}

function Timer()
{
	super.Timer();

	if( getVoteActive() )
	{
		setVoteActive( false );

		notifyAllPlayers( "Voting time has expired. Counting votes." );

		checkVotes( true );
	}
	else if( getInvokingPlayer() != none )
	{

	}
}

function bool HandleEndGame()
{
	local int i;

	Super.HandleEndGame();

	resetTempBans();

	return false;
}

// == Custom functions ==

function bool resetTempBans()
{
	local int i;

	for( i = 0; i < 32 && tempBans[ i ] != ""; i++ )
	{
		tempBans[ i ] = ""; // Forgot to put this in version 1.0B

		if( !removeBan( tempBans[ i ] ) )
		{
			log( "An error occured while attempting to remove a ban. Please check IP policies" , 'SimpleKickVote' );
			log( "IP to unban is" @ tempBans[ i ], 'SimpleKickVote' );
		}
	}
}

function bool resetVote()
{
	local int i;

	setVoteActive( false );
	votedYes = 0;
	votedNo = 0;
	setInvokingPlayer( none );
	setKickPlayer( none );

	for( i = 0; i < 32; i++ )
		alreadyVoted[ i ] = none;

	confirmVoteReset();
}

function confirmVoteReset()
{
	notifyAllPlayers( "Voting reset" );
}

/* Utility functions */

// == Ban functions ==

function bool addBan( string ipString )
{
	local int i;

	if( ipString == "" )
		return false;

	if( Level.Game != none )
	{
		if ( Level.Game.CheckIPPolicy( ipString ) )
		{
			for ( i = 0; i < 50; i++ )
			{
				if ( Level.Game.IPPolicies[ i ] == "" ) // Find empty slot
					break;
			}

			if ( i < 50 ) // Hard limit for standard ippolicies
			{
				Level.Game.IPPolicies[i] = "DENY," $ ipString;
				Level.Game.SaveConfig();

				return true;
			}
		}
	}

	return false;
}

// Could this function be exploited ?
function bool removeBan( string ipString )
{
	local int i;

	if( ipString == "" )
		return false;

	if( Level.Game != none )
	{
		if ( Level.Game.CheckIPPolicy( ipString ) )
		{
			for ( i = 0; i < 50; i++ )
			{
				if ( Level.Game.IPPolicies[ i ] == ( "DENY," $ ipString ) ) // Find banned IP
					break;
			}

			if ( i < 50 ) // Hard limit for standard ippolicies
			{
				Level.Game.IPPolicies[ i ] = "";
				Level.Game.SaveConfig();

				return true;
			}
		}
	}

	return false;
}

// == Player-specific functions ==

function private Pawn getPawnByPID( int playerID )
{
    local Pawn p;

    p = Level.PawnList;

    while ( p != none )
    {
        if( isValidPlayer( p ) )
            if( P.PlayerReplicationInfo.PlayerID == playerID ) // Pawn found
                return p; // So return reference to Pawn

        p = p.nextPawn;
    }

	return none; // No corresponding Pawn found
}

function string getPlayerIP( Pawn p )
{
	local string ip;

	if( isValidPlayer( p ) && !p.IsA( 'Bot' ) ) // Bots do not have a GetPlayerNetworkAddress() method
	{
		ip =  PlayerPawn( p ).GetPlayerNetworkAddress();
		ip = left( ip, Instr( ip, ":" ) ); // Strip network address of the port number

		return ip;
	}

	return "";
}

// Checks if we have a player
function private bool isValidPlayer( Pawn p )
{
	if( p != none &&
		( p.IsA('PlayerPawn') || p.IsA('Bot') ) &&
		p.PlayerReplicationInfo != none &&
		!p.PlayerReplicationInfo.bIsSpectator )
		return true;

	return false;
}

// == String functions ==

/* Utility functions */
function string elem( string aString, int partNumber, string separator )
{
	while( partNumber > 0)
	{
		aString = Mid( aString, InStr( aString, " " ) +1 ); // Keep going until requested part is on the left of our str

		partNumber--;
	}

	if( InStr( aString, " " ) == -1 )
		return aString; // End of string
	else
		aString = Left( aString, InStr( aString, " ")); // return requested part

	return aString;
}

defaultproperties
{
	bDebug=true
	votePercent=0.5
	voteTimeout=10.0
}
